#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **environment** (i.e., set of all external shell variables exported
to the the active Python interpreter) facilities.
'''

#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: For safety, all attempts to get, set, or unset environment variables
# should act on the low-level global "os.environ" dictionary directly rather
# than calling the high-level getenv(), setenv(), or unsetenv() functions. To
# quote official Python documentation:
#     Calling putenv() directly does not change os.environ, so it's better to
#     modify os.environ.
#
# See http://docs.python.org/library/os.html#os.environ.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ IMPORTS                            }....................
from os import environ
from betse.exceptions import BetseShellEnvException
from betse.util.type.types import type_check, MappingType, StrOrNoneTypes

# ....................{ EXCEPTIONS                         }....................
@type_check
def die_unless_var(*names: str) -> None:
    '''
    Raise an exception unless the shell environment for the current process
    defines all environment variables with the passed names.

    Parameters
    ----------
    names : tuple[str]
        Tuple of the names of all environment variables to test for.

    Raises
    ----------
    BetseShellEnvException
        If any such variable is undefined.
    '''

    # If any such variable is undefined...
    if not is_var(*names):
        # For the name of each such variable...
        for name in names:
            # If this variable is undefined, raise a human-readable exception.
            # While directly accessing this variable via "environ[name]" would
            # implicitly raise a "KeyError" exception and hence arguably
            # suffice, that exception's message is *NOT* human-readable.
            if name not in environ:
                raise BetseShellEnvException(
                    f'Environment variable "{name}" undefined by shell '
                    f'environment:\n{to_str()}'
                )

# ....................{ TESTERS                            }....................
@type_check
def is_var(*names: str) -> bool:
    '''
    ``True`` only if the shell environment for the current process defines all
    environment variables with the passed names.

    Parameters
    ----------
    names : tuple[str]
        Tuple of the names of all environment variables to test for.

    Returns
    ----------
    bool
        ``True`` only if all such variables are defined.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.iterable.mapping import maptest

    # Return true only if the names of these environment variables are all keys
    # of the corresponding dictionary global.
    return maptest.has_keys(mapping=environ, keys=names)

# ....................{ GETTERS                            }....................
def get_env() -> MappingType:
    '''
    Dictionary mapping the name of each environment variable to the string
    value of this variable.

    For safety, this dictionary is guaranteed to be a deep copy of the current
    environment and hence may be safely:

    * Modified elsewhere *without* modifying the true environment.
    * Isolate this environment to subprocesses, preventing concurrent changes
      in the environment of this process from affecting these subprocesses.
    '''

    return environ.copy()

# ....................{ GETTERS ~ var                      }....................
@type_check
def get_var(name: str) -> str:
    '''
    String value of the environment variable with the passed name if defined
    *or* raise an exception otherwise (i.e., if this variable is undefined).

    Parameters
    ----------
    name : str
        Name of this variable.

    Returns
    ----------
    str
        Value of this variable.

    Raises
    ----------
    BetseShellEnvException
        If this variable is undefined.
    '''

    # If this variable is undefined, raise an exception.
    die_unless_var(name)

    # Else, return this variable's value.
    return environ[name]


@type_check
def get_var_or_default(name: str, default: str) -> str:
    '''
    String value of the environment variable with the passed name if defined
    *or* the passed default string value otherwise.

    Parameters
    ----------
    name : str
        Name of this variable.
    default : str
        Default string value of this variable.

    Returns
    ----------
    str
        Value of this variable if defined *or* the passed default otherwise.
    '''

    return environ.get(name, default)


@type_check
def get_var_or_none(name: str) -> StrOrNoneTypes:
    '''
    String value of the environment variable with the passed name if defined
    *or* ``None`` otherwise.

    Parameters
    ----------
    name : str
        Name of this variable.

    Returns
    ----------
    StrOrNoneTypes
        Value of this variable if defined *or* ``None`` otherwise.
    '''

    return environ.get(name, None)

# ....................{ SETTERS                            }....................
@type_check
def set_var(name: str, value: str) -> None:
    '''
    Set the environment variable with the passed name to the passed value.

    Parameters
    ----------
    name : str
        Name of this variable.
    value : str
        New value of this variable.
    '''

    environ[name] = value

# ....................{ UNSETTERS                          }....................
@type_check
def unset_var(name: str) -> None:
    '''
    Unset the environment variable with the passed name (i.e., remove this
    variable from the shell environment for the current process) if defined
    *or* raise an exception otherwise (i.e., if this variable is undefined).

    Parameters
    ----------
    name : str
        Name of this variable.

    Raises
    ----------
    BetseShellEnvException
        If this variable is undefined.
    '''

    # If this variable is undefined, raise an exception.
    die_unless_var(name)

    # Unset this variable.
    unset_var_if_set(name)


@type_check
def unset_var_if_set(name: str) -> None:
    '''
    Unset the environment variable with the passed name (i.e., remove this
    variable from the shell environment for the current process) if defined
    *or* noop otherwise (i.e., if this variable is undefined).

    Parameters
    ----------
    name : str
        Name of this variable.
    '''

    # Reduce this variable to the empty string *BEFORE* unsetting this
    # variable, thus:
    #
    # * Preventing the subsequent attempt to unset this variable from raising
    #   exceptions if undefined.
    # * Handling edge-case platforms whose kernels fail to support the
    #   os.unsetenv() operation internally invoked by deleting keys from the
    #   "os.environ" dictionary (e.g., AIX). Under such platforms, deleting
    #   environment variables in Python fails to delete these variables from
    #   the environment outside of Python. Although reducing this variable to
    #   the empty string does *NOT* delete this variable, no alternatives
    #   exist.
    environ[name] = ""

    # Unset this variable.
    del environ[name]

# ....................{ CONVERTERS                         }....................
def to_str() -> str:
    '''
    Human-readable string of all environment variables.

    This string is intended to be substantially more human-readable than that
    returned by the :meth:`os._Environ.__str__` special method implicitly
    invoked by the :meth:`str.__init__` method, which compresses the entirety
    of this environment onto a single line.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.iterable.mapping import mappings

    # Return the human-readable string produced by joining on newline a
    # generator comprehension yielding the colon-delimited name and value of
    # all environment variables sorted in lexicographic order.
    #
    # Note that the "environ" object is of non-standard type "os._Environ",
    # which the pprint.pformat() fails to recognize and hence format as a
    # "dict"-compatible mapping. Ergo, passing "environ" directly to the
    # mappings.to_str_flat() function would yield a non-human-readable string.
    # While this could be ameliorated by converting "environ" to a "dict" first
    # (e.g., "mappings.to_str_flat(dict(environ))"), doing so still produces
    # less human-readable output than the current approach.
    return mappings.to_str_flat(environ)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection, PolyCollection
from betse.science import filehandling as fh
from betse.science.math.modulate import gradient_x, gradient_xr, gradient_bitmap, double_x

class WernerSim(object):

    def __init__(self, fn):

        self.sim, self.cells, self.p = fh.loadSim(fn)

    def init_sim(self):
        # define parameters:
        h = 5

        # constants for activation/inhibition of substance A:
        alpha_A = 0.1
        k_A = 1.0e-3
        D_A = 0.5e-10

        # constants for activation/inhibition of substance B:
        alpha_B = 0.4
        k_B = 2.0e-3
        D_B = 1.5e-9

        # constants for substance E:
        k_E = 2 * k_A
        D_E = 5.0e-10
        alpha_E = 0.4 * alpha_A

        # end time sequence:
        endt = 3600
        dt = 1.0e-2

        tpoints = int(endt / dt)

        #
        ttt = np.linspace(0, endt, tpoints)

        tsample = ttt[0:-1:1000]

        gx, _ = gradient_bitmap(cells.cell_centres, cells, p)

        # initial conditions for cA and cB
        # cA = np.ones(sim.cdl)*(1 + 0.25*(cells.cell_centres[:,0]/cells.cell_centres[:,0].max()))
        cA = np.ones(sim.cdl) * gx + 0.1
        cB = np.zeros(sim.cdl)
        cE = 0.1 * np.ones(sim.cdl)

    def sim_loop(self):

        for tt in ttt:

            dAt = deltaA(cA, cB, cE, D_A, alpha_A, k_A, h)
            cA += dAt * dt

            dBt = deltaB(cA, cB, cE, D_B, alpha_B, k_B, h)
            cB += dBt * dt

            dEt = deltaE(cB, D_E, alpha_E, k_E)
            cE += dEt * dt

            if tt in tsample:
                ii += 1
                print(ii, ' of ', len(tsample))
                t_time.append(1 * tt)
                cA_time.append(1 * cA)
                cB_time.append(1 * cB)
                cE_time.append(1 * cE)




    def deltaA(cA, cB, cE, D_A, alpha_A, k_A, h):
        """
        Function calculating the change for substance A in the reaction diffusion system.
        """

        # calculate the activation-inhibition term:
        termAB = (cA ** h) / (cA ** h + cB ** h)

        # calculate the extender concentration decay term:
        beta_A = k_A * cE

        # calculate the gradient of concentration A:
        gcA = (cA[cells.cell_nn_i[:, 1]] - cA[cells.cell_nn_i[:, 0]]) / (cells.nn_len)

        # calculate the diffusive flux of A:
        fluxA = -D_A * gcA

        # enforce zero flux at outer boundary:
        fluxA[cells.bflags_mems] = 0.0

        # take the divergence of the flux to obtain the net change with time:
        div_fluxA = np.dot(cells.M_sum_mems, -fluxA * cells.mem_sa) / cells.cell_vol

        # calculate the change with time for the full reaction-diffusion expression:
        dAt = alpha_A * termAB - beta_A * cA + div_fluxA

        return dAt

    def deltaB(cA, cB, cE, D_B, alpha_B, k_B, h):
        """
        Function calculating the change for substance B in the reaction diffusion system.
        """

        # calculate the activation-inhibition term:
        termAB = (cA ** h) / (cA ** h + cB ** h)

        # calculate the extender concentration decay term:
        beta_B = k_B * cE

        # calculate the gradient of concentration B:
        gcB = (cB[cells.cell_nn_i[:, 1]] - cB[cells.cell_nn_i[:, 0]]) / (cells.nn_len)

        # calculate the diffusive flux of B:
        fluxB = -D_B * gcB

        # enforce zero flux at outer boundary:
        fluxB[cells.bflags_mems] = 0.0

        # take the divergence of the flux to obtain the net change with time:
        div_fluxB = np.dot(cells.M_sum_mems, -fluxB * cells.mem_sa) / cells.cell_vol

        # calculate the change with time for the full reaction-diffusion expression:
        dBt = alpha_B * termAB - beta_B * cB + div_fluxB

        return dBt

    def deltaE(cB, D_E, alpha_E, k_E):
        """
        Function calculating the change for substance E in the reaction diffusion system.
        """

        # calculate the gradient of concentration E:
        gcE = (cE[cells.cell_nn_i[:, 1]] - cE[cells.cell_nn_i[:, 0]]) / (cells.nn_len)

        # calculate the diffusive flux of E:
        fluxE = -D_E * gcE

        # enforce zero flux at outer boundary:
        fluxE[cells.bflags_mems] = 0.0

        # take the divergence of the flux to obtain the net change with time:
        div_fluxE = np.dot(cells.M_sum_mems, -fluxE * cells.mem_sa) / cells.cell_vol

        # calculate the change with time for the full reaction-diffusion expression:
        dEt = alpha_E - k_E * cB * cE + div_fluxE

        return dEt


    def init_plots(self):

        xc = p.um * cells.cell_centres[:, 0]
        yc = p.um * cells.cell_centres[:, 1]

        xm = p.um * cells.mem_mids_flat[:, 0]
        ym = p.um * cells.mem_mids_flat[:, 1]

        xec = p.um * cells.ecm_mids[:, 0]
        yec = p.um * cells.ecm_mids[:, 1]

        xenv = p.um * cells.xypts[:, 0]
        yenv = p.um * cells.xypts[:, 1]

        xyaxis = [p.um * cells.xmin, p.um * cells.xmax, p.um * cells.ymin, p.um * cells.ymax]

        verts = p.um * cells.cell_verts

        fig = plt.figure()
        ax = plt.subplot(111)
        col = PolyCollection(verts, cmap='RdBu_r')
        col.set_array(cA_time[-1])
        ax.add_collection(col)
        fig.colorbar(col)
        plt.axis('equal')
        plt.axis(xyaxis)
        plt.show()

    def save(self):

        MMi = np.asarray([cA_time2, cB_time2, cE_time2])
        np.save('Werner_Nov/Werner_1H_2EC.npy', MMi)




#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Unit tests exercising the :mod`betse.util.type.numeric.versions` submodule.

Unit testing this submodule as critical, both because of its centrality to
sane dependency resolution *and* because of the inability of the
:mod:`setuptools`-bundled :mod:`pkg_resources` package underlying this
submodule to maintain adequate (or, *any*) backward compatibility guarantees.
'''

# ....................{ TESTS                              }....................
def test_version_comparators() -> None:
    '''
    Unit test the family of ``betse.util.type.numeric.version.is_*_than*``
    comparators (e.g.,
    :func:`betse.util.type.numeric.version.is_less_than_or_equal_to`).
    '''

    # Defer heavyweight imports.
    from betse.util.type.numeric import versions

    # Assert each comparator to behave as expected on common edge cases.
    assert versions.is_greater_than('0.6.43.410', (0, 6, 43, 409))
    assert versions.is_greater_than_or_equal_to(('3', '300', '33'), '3.300.33')
    assert versions.is_less_than(('0', '412', '454'), (0, 412, 455))
    assert versions.is_less_than_or_equal_to('23.140.692.63', '23.140.692.63')


def test_version_converters() -> None:
    '''
    Unit test the :func:`betse.util.type.numeric.version.to_comparable`
    function.
    '''

    # Defer heavyweight imports.
    from betse.util.type.numeric import versions

    # Assert this converter to behave as expected on common edge cases --
    # notably, pre-release (i.e., "alpha candidate") version numbers for the
    # PySide2 dependency required by the BETSEE downstream consumer. For
    # unknown reasons:
    # * These numbers are suffixed by "~"-prefixed labels violating PEP 440
    #   compliance. The pkg_resources package underlying this converter creates
    #   and returns a unique class of objects to represent such versions.
    # * These objects are sorted incorrectly. Versions containing tilde
    #   characters are incorrectly sorted as strictly less than versions *NOT*
    #   containing tilde characters (e.g., "5.9.0~a1" < "5.7.0").
    version_bad_new = versions.to_comparable('5.9.0~a1')
    version_bad_old = versions.to_comparable('5.6.0~a1')

    # Assert these uncompliant versions to be sorted in the correct manner.
    assert not versions.is_less_than(version_bad_new, '5.7.0')
    assert     versions.is_less_than(version_bad_old, '5.7.0')



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Integration tests exercising the public API of the public
:mod:`betse.science.wrapper` subpackage.
'''

# ....................{ IMPORTS                            }....................
from betse_test._fixture.simconf.simconfclser import SimConfTestInternal

# ....................{ TESTS                              }....................
def test_wrapper_default(betse_sim_conf: SimConfTestInternal) -> None:
    '''
    Integration test exercising default parameters accepted by the
    :meth:`betse.science.wrapper.BetseWrapper.__init__` method.

    Parameters
    ----------
    betse_sim_conf : SimConfTestInternal
        Object encapsulating a temporary simulation configuration file.
    '''

    # Defer test-specific imports.
    from betse.science.wrapper import BetseWrapper

    # # Save these changes back to the same file.
    # p.save_inplace()

    # BETSE wrapper configured by this file *AND* no other passed parameters,
    # intentionally exercising defaults set for those parameters.
    wrapper = BetseWrapper(config_filename=betse_sim_conf.p.conf_filename)

    # Run the default pipeline performing *ONLY* the initialization phase.
    wrapper.run_pipeline()


def test_wrapper_logging(
    betse_sim_conf: SimConfTestInternal,
    betse_temp_dir: 'py._path.local.LocalPath',
) -> None:
    '''
    Integration test exercising logging parameters accepted by the
    :meth:`betse.science.wrapper.BetseWrapper.__init__` method.

    Parameters
    ----------
    betse_sim_conf : SimConfTestInternal
        Object encapsulating a temporary simulation configuration file.
    betse_temp_dir : py._path.local.LocalPath
        Object encapsulating a temporary directory isolated to this test.
    '''

    # Defer test-specific imports.
    from betse.science.wrapper import BetseWrapper

    # Absolute filename of a logfile with arbitrary basename in this temporary
    # directory.
    log_file = betse_temp_dir.join('betse.log')

    # BETSE wrapper... *AND* no other passed parameters,
    # intentionally exercising defaults set for those parameters.
    wrapper = BetseWrapper(
        # Configured by this file.
        config_filename=betse_sim_conf.p.conf_filename,
        # Logging to another file.
        log_filename=str(log_file),
        # Logging *ALL* messages.
        log_level='ALL',
    )

    # Run the default pipeline performing *ONLY* the initialization phase.
    wrapper.run_pipeline()

    # Assert this logfile to have been created.
    assert log_file.check(file=1)

    # Assert this logfile to be non-empty.
    assert log_file.size() > 100



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
**Tissue profile pickers** (i.e., objects assigning a subset of all cells
matching various criteria to the corresponding tissue profile) classes.
'''

# ....................{ IMPORTS                           }....................
from abc import ABCMeta, abstractmethod
from betse.exceptions import BetseSimConfException
from betse.lib.numpy import nparray
from betse.science.math import toolbox
# from betse.util.io.log import logs
from betse.util.type.types import type_check, NumericSimpleTypes, SequenceTypes
import numpy as np

# ....................{ SUPERCLASS                        }....................
class TissuePickerABC(object, metaclass=ABCMeta):
    '''
    Abstract base class of all **tissue profile picker** (i.e., object
    assigning a subset of all cells matching some criteria to the corresponding
    tissue profile) subclasses.

    Typical criteria for matching cells includes explicit cell indexing,
    randomized cell selection, and image-defined spatial cell location.
    '''

    # ..................{ PICKERS                           }..................
    @abstractmethod
    def pick_cells(
        self,
        cells: 'betse.science.cells.Cells',
        p:     'betse.science.parameters.Parameters',
    ) -> SequenceTypes:
        '''
        One-dimensional Numpy array of the indices of all cells in the passed
        cell cluster selected by this tissue picker, ignoring extracellular
        spaces.

        Parameters
        ----------
        cells : Cells
            Current cell cluster.
        p : Parameters
            Current simulation configuration.

        Returns
        ----------
        SequenceTypes
            One-dimensional Numpy array of the indices of all such cells.
        '''

        pass


    #FIXME: Rather awkward. We actually end up passing the two Numpy arrays
    #embedded in the tuple returned by this method together as separate
    #parameters to other methods throughout the codebase (e.g., the
    #mod_after_cut_event() and remove_cells() methods). Instead:
    #
    #* Define a new simple "TissuePicked" class at the top of this submodule
    #  (e.g., via the betse.util.type.iterable.tuples.make_named_subclass()
    #  class factory function).
    #* This class should provide only the following public fields:
    #  * "cells_index".
    #  * "membs_index".
    #* Refactor this method to create and return an instance of that class.
    #* Refactor all other methods requiring this pair of Numpy arrays to accept
    #  an instance of that class.
    #* Refactor the following pair of instance variables into a single
    #  instance variable whose value is an instance of this class:
    #  * "Simulator.target_inds_cell_o".
    #  * "Simulator.target_inds_mem_o".
    @type_check
    def pick_cells_and_mems(
        self,
        cells: 'betse.science.cells.Cells',
        p:     'betse.science.parameters.Parameters',
    ) -> tuple:
        '''
        2-tuple ``(cells_index, mems_index)`` of one-dimensional Numpy arrays
        of the indices of both all cells *and* cell membranes in the passed
        cell cluster selected by this tissue picker, ignoring extracellular
        spaces.

        By default, this method returns the array returned by the subclass
        implementation of the abstract :meth:`pick_cells` method, mapped from
        cell to cell membrane indices.

        Parameters
        ----------
        cells : Cells
            Current cell cluster.
        p : Parameters
            Current simulation configuration.

        Returns
        ----------
        (ndarray, ndarray)
            2-tuple ``(cells_index, mems_index)``, where:
            * ``cells_index`` is the one-dimensional Numpy array of the indices
              of all subclass-selected cells.
            * ``mems_index`` is the one-dimensional Numpy array of the indices
              of all subclass-selected cell membranes.
        '''

        # One-dimensional Numpy array of the indices of subclass-selected cells.
        cells_index = self.pick_cells(cells, p)

        #FIXME: Ideally, this array would be trivially defined as follows:
        #    mems_index = cells.cell_to_mems[cells_index].flatten()
        #Unfortunately, the two-dimensional Numpy array "cells.cell_to_mems"
        #actually appears to be a one-dimensional array of lists -- which is a
        #bit bizarro-world. To compaund matters, this array's "dtype" is
        #"object" (presumably, because it contains Python lists) rather than
        #the "dtype" of "int" that one might expect. Since this is the case,
        #it's infeasible to even coerce the temporary one-dimensional Numpy
        #array "cells.cell_to_mems[cells_index]" into a two-dimensional array
        #of ints by calling ndarray.astype(int). Frankly, it's all a bit beyond
        #me; until this core issue is resolved, however, the current inelegant
        #and inefficient approach remains.

        # One-dimensional Numpy array of the indices of subclass-selected cell
        # membranes mapped from the array of cell indices.
        # logs.log_debug('cells_index: %r', cells_index)
        mems_index = cells.cell_to_mems[cells_index]
        mems_index, _, _ = toolbox.flatten(mems_index)
        mems_index = nparray.from_iterable(mems_index)

        # Return these arrays.
        return cells_index, mems_index

# ....................{ SUBCLASSES                        }....................
class TissuePickerAll(TissuePickerABC):
    '''
    All-inclusive tissue picker, unconditionally matching *all* cells in the
    current cell cluster.
    '''

    # ..................{ PICKERS                           }..................
    @type_check
    def pick_cells(
        self,
        cells: 'betse.science.cells.Cells',
        p:     'betse.science.parameters.Parameters',
    ) -> SequenceTypes:

        return cells.cell_i


class TissuePickerColor(TissuePickerABC):
    '''
    Vector image-based tissue picker, matching all cells in the current cell
    cluster whose cell centres are simple circles with a given fill color of a
    vector image (defined by the ``cells from svg`` setting in the current
    YAML-formatted simulation configuration file).

    Attributes
    ----------
    cells_color : str
        **Hexadecimal-formatted color** (i.e., string of six hexadecimal digits
        specifying this color's red, green, and blue components) of all circles
        within this vector image to be selected as cell centres.
    '''

    # ..................{ INITIALIZERS                      }..................
    @type_check
    def __init__(self, cells_color: str) -> None:
        '''
        Initialize this tissue picker.

        Parameters
        ----------
        cells_color : str
            **Hexadecimal-formatted color** (i.e., string of six hexadecimal
            digits, specifying this color's red, green, and blue components) of
            all circles within this vector image to be selected as cell
            centres.
        '''

        self.cells_color = cells_color

    # ..................{ PICKERS                           }..................
    @type_check
    def pick_cells(
        self,
        cells: 'betse.science.cells.Cells',
        p:     'betse.science.parameters.Parameters',
    ) -> SequenceTypes:

        # If no cell cluster SVG is enabled, raise an exception.
        if cells.seed_fills is None:
            raise BetseSimConfException(
                'Color-based cell targets type requires '
                'cell cluster SVG to be enabled.')

        # Search for cells matching the tissue profile color.
        selected_cells = []

        for i, fi in enumerate(cells.seed_fills):
            if fi == self.cells_color:
                selected_cells.append(i)

        return selected_cells


class TissuePickerIndices(TissuePickerABC):
    '''
    Cell indices-based tissue picker, matching all cells in the current cell
    cluster whose indices are defined by a given sequence.

    Attributes
    ----------
    cells_index : SequenceTypes
        Sequence of the indices of all cells to match.
    '''

    # ..................{ INITIALIZERS                      }..................
    @type_check
    def __init__(self, cells_index: SequenceTypes) -> None:
        '''
        Initialize this tissue picker.

        Parameters
        ----------
        cells_index : SequenceTypes
            Sequence of the indices of all cells to match.
        '''

        self.cells_index = cells_index

    # ..................{ PICKERS                           }..................
    @type_check
    def pick_cells(
        self,
        cells: 'betse.science.cells.Cells',
        p:     'betse.science.parameters.Parameters',
    ) -> SequenceTypes:

        return self.cells_index


class TissuePickerPercent(TissuePickerABC):
    '''
    Randomized cell picker, randomly matching a given percentage of all cells
    in the current cell cluster.

    Attributes
    ----------
    cells_percent : NumericSimpleTypes
        **Percentage** (i.e., number in the range ``[0.0, 100.0]``) of the
        total cell population to randomly match.
    '''

    # ..................{ INITIALIZERS                      }..................
    @type_check
    def __init__(self, cells_percent: NumericSimpleTypes) -> None:
        '''
        Initialize this tissue picker.

        Parameters
        ----------
        cells_percent : NumericSimpleTypes
            **Percentage** (i.e., number in the range ``[0.0, 100.0]``) of the
            total cell population to randomly match.
        '''

        # If this is not a valid percentage, raise an exception. This is
        # important enough to always test rather than defer to assertions.
        if not 0.0 <= cells_percent <= 100.0:
            raise BetseSimConfException(
                'Percentage {} not in range [0, 100].'.format(cells_percent))

        # Classify this parameter.
        self.cells_percent = cells_percent

    # ..................{ PICKERS                           }..................
    @type_check
    def pick_cells(
        self,
        cells: 'betse.science.cells.Cells',
        p:     'betse.science.parameters.Parameters',
    ) -> SequenceTypes:

        # Total number of cells in this cluster.
        data_length = len(cells.cell_i)

        # Total number of cells to randomly select from this cluster.
        data_fraction = int((self.cells_percent/100)*data_length)

        cell_i_copy = cells.cell_i[:]
        np.random.shuffle(cell_i_copy)

        # For simplicity, non-randomly select the indices of the first
        # "data_fraction"-th cells in this cluster. While technically
        # non-random, this simplistic approach currently suffices.
        return [cell_i_copy[x] for x in range(0, data_fraction)]



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
**Simulation configuration enumerations** (i.e., :class:`enum.Enum` subclasses
required to validate typed data in simulation configuration files).
'''

# ....................{ IMPORTS                           }....................
from betse.util.type import enums

# ....................{ ENUMS ~ cell                      }....................
CellLatticeType = enums.make_enum(
    class_name='CellLatticeType',
    member_names=('HEX', 'SQUARE',),
    doc='''
Enumeration of all supported types of **base cell lattices** (i.e., uniform
grid to which cells are situated *before* random lattice disorder is applied).

Attributes
----------
HEX : enum
    Hexagonal base cell lattice, situating cells along a hexagonal grid.
SQUARE : enum
    Rectilinear base cell lattice, situating cells along a square grid.
''')

# ....................{ ENUMS ~ cells                     }....................
CellsPickerType = enums.make_enum(
    class_name='CellsPickerType',
    member_names=('ALL', 'COLOR', 'IMAGE', 'INDICES', 'PERCENT',),
    doc='''
Enumeration of all supported types of **tissue profile pickers** (i.e., objects
assigning a subset of all cells matching some criteria to the corresponding
tissue profile).

Attributes
----------
ALL : enum
    All-inclusive tissue picker, unconditionally matching *all* cells.
COLOR : enum
    Vector image-based tissue picker, matching all cells whose cell centres are
    simple circles with a given fill color of a given vector (i.e., scalable
    vector graphics (SVG)-formatted) image file.
IMAGE : enum
    Raster image-based tissue picker, matching all cells whose cell centres are
    the pure-black pixels of a given raster (i.e., bitmap) image file.
INDICES : enum
    Cell indices-based tissue picker, matching all cells whose indices are
    defined by a given sequence.
PERCENT : enum
    Randomized cell picker, randomly matching a given percentage of all cells.
''')

# ....................{ ENUMS ~ grn                       }....................
GrnUnpicklePhaseType = enums.make_enum(
    class_name='GrnUnpicklePhaseType',
    member_names=('SEED', 'INIT', 'SIM',),
    doc='''
Enumeration of all supported types of gene regulatory network (GRN)-specific
**unpickle simulation phases** (i.e., previously pickled simulation phase to
unpickle as the computational basis for the current network to be run by the
``betse sim-grn`` subcommand).

Attributes
----------
SEED : enum
    Unpickled seed phase, running this network on the results of a previously
    run ``betse seed`` subcommand and hence an uninitialized and unsimulated
    cell cluster (i.e., neither initialized to steady-state concentrations
    *nor* simulated with non-steady-state bioelectrical phenomena).
INIT : enum
    Unpickled initialization phase, running this network on the results of a
    previously run ``betse init`` subcommand and hence an initialized but
    unsimulated cell cluster.
SIM : enum
    Unpickled simulation phase, running this network on the results of a
    previously run ``betse sim`` subcommand and hence an initialized and
    simulated cell cluster.
''')

# ....................{ ENUMS ~ ion                       }....................
IonProfileType = enums.make_enum(
    class_name='IonProfileType',
    member_names=('BASIC', 'BASIC_CA', 'MAMMAL', 'AMPHIBIAN', 'CUSTOM',),
    doc='''
Enumeration of all supported types of **ion profiles** (i.e., predefined sets
of all extracellular and cytosolic ions enabled by this simulation).

Note that *all* ion profiles unconditionally enable an unidentified
charge-balance anion denoted M-, as required for both simulation stability and
correctness.

Attributes
----------
BASIC : enum
    Ion profile enabling M-, Na+, K+, and proteins- ions. This profile is the
    proper subset of all other predefined ion profiles.
BASIC_CA : enum
    Ion profile enabling M-, Na+, K+, proteins-, and Ca2+ ions. This profile is
    the superset of the :attr:`BASIC` profile enabling Ca2+ ions.
MAMMAL : enum
    Ion profile enabling M-, Na+, K+, proteins-, Ca2+, Cl-, and H+ ions
    expressed in amniotic environmental concentrations, principally intended
    for mammalian cell clusters. This profile is the superset of the
    :attr:`BASIC_CA` profile enabling Cl- and H+ ions.
AMPHIBIAN : enum
    Ion profile enabling M-, Na+, K+, proteins-, Ca2+, Cl-, and H+ ions
    expressed in aquatic environmental concentrations, principally intended for
    amphibian cell clusters. This profile enables the same ions as the
    :attr:`MAMMAL` profile -- albeit in differing concentrations.
CUSTOM : enum
    User-defined ion profile. See the
    :attr:`betse.science.parameters.Parameters.ions_custom` variable.
''')

# ....................{ ENUMS ~ solver                    }....................
SolverType = enums.make_enum(
    class_name='SolverType',
    #FIXME: Replace this set of members with the following when sufficient time
    #permits to properly implement the DEC-based solver referred to as "MAX".
    member_names=('FULL', 'FAST',),
    # member_names=('FULL', 'FAST', 'MAX',),
    doc='''
Enumeration of all supported types of **simulation solvers** (i.e.,
computational technique iteratively computing each time step of a simulation).

Attributes
----------
FAST : enum
    Inaccurate space- and time-efficient (i.e., fast) solver based on the
    well-known equivalent circuit formalism. This solver produces draft-quality
    results, analogizing real-world biological systems to electronic circuits.
    While integrated with gene regulatory networks (GRNs), this solver *cannot*
    by definition simulate the following phenomena simulated by the
    :attr:`FULL` and :attr:`MAX` solvers:

    * Bioelectric fields or currents.
    * Extracellular voltages or voltage polarities.
    * Ion concentrations.

    If this solver is selected, these phenomena are silently ignored even if
    otherwise enabled by the current simulation configuration.
FULL : enum
    Accurate space- and time-inefficient (i.e., slow) solver based on a unique
    amalgam of discrete exterior calculus (DEC) and finite volume techniques.
    This solver produces publication-quality results, simulating most
    bioelectrical phenomena expected of real-world biological systems.
MAX : enum
    **(Prospective)** Extremely accurate, space- and time-efficient (i.e.,
    fast) solver based exclusively on discrete exterior calculus (DEC)
    techniques reproducing Maxwell's equations. This solver produces
    publication-quality results, comprehensively simulating *all* bioelectrical
    phenomena expected of real-world biological systems.
''')



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
**Post-simulation cell cluster plot pipelines** (i.e., pipelines plotting
simulated data of all cells in the cell cluster).
'''

# ....................{ IMPORTS                            }....................
import numpy as np
from betse.science.config.export.visual.confexpvisplot import (
    SimConfExportPlotCells)
from betse.science.math import mathunit
from betse.science.phase.phasecls import SimPhase
from betse.science.phase.require import phasereqs
from betse.science.pipe.export.plot.pipeexpplotabc import SimPipeExportPlotABC
from betse.science.pipe.piperun import piperunner
from betse.science.visual.plot import plotutil
from betse.util.io.log import logs
from betse.util.type.iterable import iterget
from betse.util.type.descriptor.descs import classproperty_readonly
from betse.util.type.types import type_check, SequenceTypes
from collections import OrderedDict
from matplotlib import pyplot
from matplotlib.collections import LineCollection, PolyCollection

# ....................{ SUBCLASSES                         }....................
class SimPipeExportPlotCells(SimPipeExportPlotABC):
    '''
    **Post-simulation cell cluster plot pipeline** (i.e., object iteratively
    displaying and/or saving all plots depicting all cells of the cell cluster,
    produced after initialization and simulation as specified by the current
    simulation configuration).
    '''

    # ..................{ SUPERCLASS ~ properties            }..................
    @classproperty_readonly
    def _NOUN_SINGULAR(cls) -> str:
        return 'cell cluster plot'

    # ..................{ SUPERCLASS                         }..................
    @type_check
    def iter_runners_conf(self, phase: SimPhase) -> SequenceTypes:
        return phase.p.plot.plots_cells_after_sim

    # ..................{ EXPORTERS ~ channel                }..................
    # @piperunner(
    #     categories=('Ion Channel', 'Density Factor',),
    #     requirements={phasereqs.ELECTROOSMOSIS,},
    # )
    # def export_channel_density(self, conf: SimConfExportPlotCells) -> None:
    #     '''
    #     Plot all cell membrane ion channel density factors for the cell cluster
    #     at the last time step.
    #     '''
    #
    #     # Prepare to export the current plot.
    #     self._export_prep()
    #
    #     plotutil.plotMemData(
    #         phase.cells,
    #         phase.p,
    #         zdata=phase.sim.rho_channel,
    #         clrmap=phase.p.default_cm,
    #     )
    #     pyplot.xlabel('Spatial Dimension [um]')
    #     pyplot.ylabel('Spatial Dimension [um]')
    #     pyplot.title('Membrane ion channel density factor')
    #
    #     # Export this plot to disk and/or display.
    #     self._export(basename='final_channels_2D')

    # ..................{ EXPORTERS ~ cluster                }..................
    # This exporter is solver- and feature-agnostic.

    @piperunner(categories=('Cell Cluster', 'Mask',))
    def export_cluster_mask(
        self, phase: SimPhase, conf: SimConfExportPlotCells) -> None:
        '''
        Plot the **cell cluster image mask** (i.e., user-defined image whose
        pure-black pixels exactly correspond to the shape of the cell cluster).

        This plot is irrespective of time step.
        '''

        # Prepare to export the current plot.
        self._export_prep(phase)

        fig = pyplot.figure()
        pyplot.imshow(
            phase.cells.maskM,
            origin='lower',
            extent=phase.cache.upscaled.extent,
            cmap=phase.p.background_cm,
        )
        pyplot.colorbar()
        pyplot.title('Cluster Masking Matrix')

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='cluster_mask')

    # ..................{ EXPORTERS ~ cluster : tissue       }..................
    # This exporter is solver- and feature-agnostic.
    @piperunner(categories=('Cell Cluster', 'Tissue and Cut Profiles',))
    def export_tissue_cuts(
        self, phase: SimPhase, conf: SimConfExportPlotCells) -> None:
        '''
        Plot a **tissue and cut profile tessellation** (i.e., tiled mosaic of
        all cells spatially subdivided into tissue and cut profile regions such
        that all cells in the same region share the same arbitrary color) for
        the cell cluster.

        This plot is irrespective of time step.
        '''

        # Prepare to export the current plot.
        self._export_prep(phase)

        # Localize frequently accessed fields for convenience.
        p = phase.p
        dyna = phase.dyna
        cells = phase.cells
        colormap = phase.p.background_cm

        col_dic = {}
        cb_ticks = []
        cb_tick_labels = []

        # Ordered dictionary mapping from the name of each tissue and cut
        # profile to a one-dimensional Numpy array of the 0-based indices of
        # all cells owned by this profile.
        #
        # Note that order is absolutely significant. The first tissue profile
        # is a pseudo-tissue defining the cell cluster itself. If order were
        # *NOT* preserved here, this tissue would be assigned an arbitrary
        # z-order, in which case all tissues assigned smaller z-orders would be
        # entirely obscured by that pseudo-tissue covering the cell cluster.
        profile_name_to_cells_index = OrderedDict()

        fig = pyplot.figure()
        ax = pyplot.subplot(111)

        # For the name and object encapsulating each tissue profile...
        for tissue_name, _ in dyna.tissue_name_to_profile.items():
            # One-dimensional Numpy array of the indices of all tissue cells.
            tissue_cells_index = dyna.cell_target_inds[tissue_name]

            # If this tissue contains no cells, skip to the next tissue.
            if not len(tissue_cells_index):
                logs.log_warning('Tissue "%s" contains no cells.', tissue_name)
                continue

            # Else, tissue contains one or more cells. Map this tissue to these
            # indices (for subsequent lookup).
            profile_name_to_cells_index[tissue_name] = tissue_cells_index

        #FIXME: The "p.plot_cutlines" boolean is only ever leveraged here and
        #thus is arguably extraneous. Consider refactoring as follows:
        #
        #* Remove the "p.plot_cutlines" boolean and corresponding
        #  YAML-formetted default configuration option.
        #* Split the existing "tissue_cuts" plot type in the "cell cluster
        #  pipeline" into the following two types:
        #  * "tissue", unconditionally plotting *ONLY* tissue profiles.
        #  * "tissue_cuts", unconditionally plotting both tissue and cut
        #    profiles.
        #* Define a new private _export_profiles() method as follows:
        #      @type_check
        #      _export_profiles(
        #          self,
        #          conf: SimConfExportPlotCells,
        #          is_tissue: bool,
        #          is_cuts: bool
        #       ) -> None:
        #* Implement export_tissue() to call _export_profiles() as:
        #    self._export_profiles(
        #        self,
        #        conf=conf,
        #        is_tissue=True,
        #        is_cuts=False,
        #    )
        #* Implement export_tissue_cuts() similarly.

        # If plotting cut profiles as well *AND* the cutting event is
        # enabled...
        if p.plot_cutlines and dyna.event_cut is not None:
            # For the name and object encapsulating each cut profile...
            for cut_name, cut_profile in dyna.cut_name_to_profile.items():
                # Map this cut to the indices of all cells cut by this profile.
                profile_name_to_cells_index[cut_name] = (
                    cut_profile.picker.pick_cells(cells=cells, p=p))

        # Minimum and maximum 0-based integers uniquely identifying the first
        # and last tissue and cut profile (respoctively), localized for
        # ordering purposes in the colorbar legend.
        profile_zorder = 0
        profile_zorder_max = len(profile_name_to_cells_index)

        # For the name and one-dimensional Numpy array of the 0-based indices
        # of all cells in each tissue and/or cut profile...
        for profile_name, profile_cells_index in (
            profile_name_to_cells_index.items()):
            # logs.log_debug('Plotting tissue "%s"...', profile_name)
            profile_zorder += 1

            profile_points = mathunit.upscale_coordinates(
                cells.cell_verts[profile_cells_index])

            z = np.zeros(len(profile_points))
            z[:] = profile_zorder

            col_dic[profile_name] = PolyCollection(
                profile_points, array=z, cmap=colormap, edgecolors='none')
            col_dic[profile_name].set_clim(0, profile_zorder_max)

            # col_dic[profile_name].set_alpha(0.8)
            col_dic[profile_name].set_zorder(profile_zorder)
            ax.add_collection(col_dic[profile_name])

            # Add this profile name to the colour legend.
            cb_ticks.append(profile_zorder)
            cb_tick_labels.append(profile_name)

        # logs.log_debug('Plotting colorbar ticks: %r', cb_ticks)
        # logs.log_debug('Plotting colorbar tick labels: %r', cb_tick_labels)

        ax_cb = None
        if dyna.tissue_name_to_profile:
            # Name of the first tissue profile.
            tissue_first_name = iterget.get_item_first(
                dyna.tissue_name_to_profile.keys())

            # Color mappable associated with this tissue profile, guaranteed in
            # this case to be a "PolyCollection" instance.
            tissue_first_mappable = col_dic[tissue_first_name]

            ax_cb = fig.colorbar(tissue_first_mappable, ax=ax, ticks=cb_ticks)
            ax_cb.ax.set_yticklabels(cb_tick_labels)

        if p.visual.is_show_cell_indices:
            for i, cll in enumerate(cells.cell_centres):
                ax.text(
                    p.um*cll[0], p.um*cll[1], i,
                    ha='center', va='center', zorder=20)

        ax.set_xlabel('Spatial Distance [um]')
        ax.set_ylabel('Spatial Distance [um]')
        ax.set_title('Cell Cluster')

        ax.axis('equal')
        ax.axis(phase.cache.upscaled.extent)

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='cluster_mosaic')

    # ..................{ EXPORTERS ~ current                }..................
    @piperunner(
        categories=('Current Density', 'Intracellular',),
        requirements=phasereqs.ELECTRIC_CURRENT,
    )
    def export_currents_intra(
        self, phase: SimPhase, conf: SimConfExportPlotCells) -> None:
        '''
        Plot all intracellular current densities for the cell cluster at the
        last time step.
        '''

        # Prepare to export the current plot.
        self._export_prep(phase)

        figI, axI, cbI = plotutil.plotStreamField(
            100*phase.sim.J_cell_x,
            100*phase.sim.J_cell_y,
            phase.cells,
            phase.p,
            plot_ecm=False,
            title='Intracellular Current Density',
            cb_title='Current Density [uA/cm2]',
            show_cells=False,
            colorAutoscale=conf.is_color_autoscaled,
            minColor=conf.color_min,
            maxColor=conf.color_max,
        )

        axI.set_xlabel('Spatial distance [um]')
        axI.set_ylabel('Spatial distance [um]')
        cbI.set_label('Current Density [uA/cm2]')

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='Final_Current_gj')


    @piperunner(
        categories=('Current Density', 'Extracellular',),
        requirements=phasereqs.ELECTRIC_CURRENT_EXTRA,
    )
    def export_currents_extra(
        self, phase: SimPhase, conf: SimConfExportPlotCells) -> None:
        '''
        Plot all extracellular current densities for the environment at the
        last time step.
        '''

        # Prepare to export the current plot.
        self._export_prep(phase)

        figI2, axI2, cbI2 = plotutil.plotStreamField(
            100*phase.sim.J_env_x,
            100*phase.sim.J_env_y,
            phase.cells,
            phase.p,
            plot_ecm=True,
            title='Extracellular Current Density',
            cb_title='Current Density [uA/cm2]',
            show_cells=False,
            colorAutoscale=conf.is_color_autoscaled,
            minColor=conf.color_min,
            maxColor=conf.color_max,
        )

        axI2.set_xlabel('Spatial distance [um]')
        axI2.set_ylabel('Spatial distance [um]')
        cbI2.set_label('Extracellular Current Density [uA/cm2]')

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='Final_Current_extracellular')

    # ..................{ EXPORTERS ~ deform                 }..................
    @piperunner(
        categories=('Deformation', 'Total',),
        requirements=phasereqs.DEFORM,
    )
    def export_deform_total(
        self, phase: SimPhase, conf: SimConfExportPlotCells) -> None:
        '''
        Plot all **total cellular displacements** (i.e., summations of all
        cellular deformations due to galvanotropic and osmotic pressure body
        forces) for the cell cluster at the last time step.
        '''

        # Prepare to export the current plot.
        self._export_prep(phase)

        plotutil.plotStreamField(
            phase.p.um*phase.sim.dx_cell_time[-1],
            phase.p.um*phase.sim.dy_cell_time[-1],
            phase.cells, phase.p,
            plot_ecm=False,
            title='Final Displacement of Cell Collective',
            cb_title='Displacement [um]',
            show_cells=phase.p.showCells,
            colorAutoscale=conf.is_color_autoscaled,
            minColor=conf.color_min,
            maxColor=conf.color_max,
        )

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='final_displacement_2D')

    # ..................{ EXPORTERS ~ diffusion              }..................
    @piperunner(
        categories=('Diffusion Weights', 'Extracellular',),
        requirements=phasereqs.ECM,
    )
    def export_diffusion_extra(
        self, phase: SimPhase, conf: SimConfExportPlotCells) -> None:
        '''
        Plot all logarithm-scaled extracellular diffusion weights for the
        environment at the last time step.
        '''

        # Prepare to export the electric plot.
        self._export_prep(phase)

        fig = pyplot.figure()
        ax99 = pyplot.subplot(111)
        pyplot.imshow(
            np.log10(phase.sim.D_env_weight.reshape(
                phase.cells.X.shape)),
            origin='lower',
            extent=phase.cache.upscaled.extent,
            cmap=phase.p.background_cm,
        )
        pyplot.colorbar()

        cell_edges_flat = mathunit.upscale_coordinates(
            phase.cells.mem_edges_flat)
        coll = LineCollection(cell_edges_flat, colors='k')
        coll.set_alpha(1.0)
        ax99.add_collection(coll)

        pyplot.title('Logarithm of Environmental Diffusion Weight Matrix')

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='env_diffusion_weights')

    # ..................{ EXPORTERS ~ electric               }..................
    @piperunner(
        categories=('Electric Field', 'Intracellular',),
        requirements=phasereqs.ELECTRIC_FIELD,
    )
    def export_electric_intra(
        self, phase: SimPhase, conf: SimConfExportPlotCells) -> None:
        '''
        Plot all intracellular electric field lines for the cell cluster at the
        last time step.
        '''

        # Prepare to export the electric plot.
        self._export_prep(phase)

        plotutil.plotVectField(
            phase.sim.E_cell_x,
            phase.sim.E_cell_y,
            phase.cells,
            phase.p,
            plot_ecm=False,
            title='Final Electric Field',
            cb_title='Electric Field [V/m]',
            colorAutoscale=conf.is_color_autoscaled,
            minColor=conf.color_min,
            maxColor=conf.color_max,
        )

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='Final_Electric_Field_GJ')


    @piperunner(
        categories=('Electric Field', 'Extracellular',),
        requirements=phasereqs.ELECTRIC_FIELD_EXTRA,
    )
    def export_electric_extra(
        self, phase: SimPhase, conf: SimConfExportPlotCells) -> None:
        '''
        Plot all extracellular electric field lines for the environment at the
        last time step.
        '''

        # Prepare to export the electric plot.
        self._export_prep(phase)

        plotutil.plotVectField(
            phase.sim.E_env_x,
            phase.sim.E_env_y,
            phase.cells,
            phase.p,
            plot_ecm=True,
            title='Final Electric Field',
            cb_title='Electric Field [V/m]',
            colorAutoscale=conf.is_color_autoscaled,
            minColor=conf.color_min,
            maxColor=conf.color_max,
        )

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='Final_Electric_Field_ECM')

    # ..................{ EXPORTERS ~ fluid                  }..................
    @piperunner(
        categories=('Fluid Flow', 'Intracellular',),
        requirements=phasereqs.FLUID,
    )
    def export_fluid_intra(
        self, phase: SimPhase, conf: SimConfExportPlotCells) -> None:
        '''
        Plot all intracellular fluid flow field lines for the cell cluster at
        the last time step.
        '''

        # Prepare to export the current plot.
        self._export_prep(phase)

        plotutil.plotStreamField(
            1e6*phase.sim.u_cells_x,
            1e6*phase.sim.u_cells_y,
            phase.cells, phase.p,
            plot_ecm=False,
            title='Final Fluid Velocity in Cell Collective',
            cb_title='Velocity [um/s]',
            colorAutoscale=conf.is_color_autoscaled,
            minColor=conf.color_min,
            maxColor=conf.color_max,
        )

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='final_vel_2D_gj')


    @piperunner(
        categories=('Fluid Flow', 'Extracellular',),
        requirements=phasereqs.FLUID_EXTRA,
    )
    def export_fluid_extra(
        self, phase: SimPhase, conf: SimConfExportPlotCells) -> None:
        '''
        Plot all extracellular fluid flow field lines for the environment at
        the last time step.
        '''

        # Prepare to export the current plot.
        self._export_prep(phase)

        plotutil.plotStreamField(
            1e9*phase.sim.u_env_x,
            1e9*phase.sim.u_env_y,
            phase.cells, phase.p,
            plot_ecm=True,
            title='Final Fluid Velocity in Environment',
            cb_title='Velocity [nm/s]',
            colorAutoscale=conf.is_color_autoscaled,
            minColor=conf.color_min,
            maxColor=conf.color_max,
        )

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='final_vel_2D_env')

    # ..................{ EXPORTERS ~ ion : calcium          }..................
    @piperunner(
        categories=('Ion Concentration', 'Calcium', 'Intracellular',),
        requirements=phasereqs.ION_CALCIUM,
    )
    def export_ion_calcium_intra(
        self, phase: SimPhase, conf: SimConfExportPlotCells) -> None:
        '''
        Plot all intracellular calcium (i.e., Ca2+) ion concentrations for the
        cell cluster at the last time step.
        '''

        # Prepare to export the current plot.
        self._export_prep(phase)

        figCa, axCa, cbCa = plotutil.plotPolyData(
            phase.sim, phase.cells, phase.p,
            zdata=phase.sim.cc_time[-1][phase.sim.iCa]*1e6,
            number_cells=phase.p.visual.is_show_cell_indices,
            clrmap=phase.p.default_cm,
            clrAutoscale=conf.is_color_autoscaled,
            clrMin=conf.color_min,
            clrMax=conf.color_max,
        )

        axCa.set_title('Final cytosolic Ca2+')
        axCa.set_xlabel('Spatial distance [um]')
        axCa.set_ylabel('Spatial distance [um]')
        cbCa.set_label('Concentration nmol/L')

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='final_Ca_2D')


    @piperunner(
        categories=('Ion Concentration', 'Calcium', 'Extracellular',),
        requirements=phasereqs.ION_CALCIUM_EXTRA,
    )
    def export_ion_calcium_extra(
        self, phase: SimPhase, conf: SimConfExportPlotCells) -> None:
        '''
        Plot all extracellular calcium (i.e., Ca2+) ion concentrations for the
        environment at the last time step.
        '''

        # Prepare to export the current plot.
        self._export_prep(phase)

        cc_Ca = phase.sim.cc_env[phase.sim.iCa].reshape(
            phase.cells.X.shape)

        pyplot.figure()
        pyplot.imshow(
            cc_Ca,
            origin='lower',
            extent=phase.cache.upscaled.extent,
            cmap=phase.p.default_cm,
        )
        pyplot.colorbar()
        pyplot.title('Environmental Calcium [mmol/L]')

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='Final_environmental_calcium')

    # ..................{ EXPORTERS ~ gap junction           }..................
    # These exporters are solver- and feature-agnostic.

    @piperunner(categories=('Gap Junction', 'Connectivity Network',))
    def export_gj_connectivity(
        self, phase: SimPhase, conf: SimConfExportPlotCells) -> None:
        '''
        Plot the **gap junction connectivity network** (i.e., graph of all gap
        junctions connecting cell membranes) for the cell cluster.

        This plot is irrespective of time step.
        '''

        # Prepare to export the current plot.
        self._export_prep(phase)

        fig = pyplot.figure()
        ax_x = pyplot.subplot(111)

        if phase.p.showCells:
            base_points = mathunit.upscale_coordinates(
                phase.cells.cell_verts)
            col_cells = PolyCollection(
                base_points, facecolors='k', edgecolors='none')
            col_cells.set_alpha(0.3)
            ax_x.add_collection(col_cells)

        connects = mathunit.upscale_coordinates(phase.cells.nn_edges)
        collection = LineCollection(connects, linewidths=1.0, color='b')
        ax_x.add_collection(collection)
        pyplot.axis('equal')
        pyplot.axis(phase.cache.upscaled.extent)

        ax_x.set_xlabel('Spatial x [um]')
        ax_x.set_ylabel('Spatial y [um')
        ax_x.set_title('Cell Connectivity Network')

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='gj_connectivity_network')


    @piperunner(categories=('Gap Junction', 'Relative Permeability',))
    def export_gj_permeability(
        self, phase: SimPhase, conf: SimConfExportPlotCells) -> None:
        '''
        Plot all **gap junction connectivity states** (i.e., relative
        permeabilities of all gap junctions connecting cell membranes) for the
        cell cluster at the last time step.
        '''

        # Prepare to export the current plot.
        self._export_prep(phase)

        fig_x = pyplot.figure()
        ax_x = pyplot.subplot(111)

        con_segs = phase.cells.nn_edges
        connects = phase.p.um*np.asarray(con_segs)
        collection = LineCollection(
            connects,
            array=phase.sim.gjopen,
            cmap=phase.p.background_cm,
            linewidths=2.0,
        )
        # collection.set_clim(0, 1)

        ax_x.add_collection(collection)
        cb = fig_x.colorbar(collection)
        pyplot.axis('equal')
        pyplot.axis(phase.cache.upscaled.extent)

        cb.set_label('Relative Permeability')
        ax_x.set_xlabel('Spatial x [um]')
        ax_x.set_ylabel('Spatial y [um')
        ax_x.set_title('Final Gap Junction Relative Permeability')

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='final_gjState')

    # # ..................{ EXPORTERS ~ microtubule          }..................
    #FIXME: Restore after proper support for microtubules is implemented.
    # @piperunner(
    #     categories=('Microtubule', 'Coherence',),
    #     requirements=phasereqs.MICROTUBULE,
    # )
    # def export_microtubule(
    #     self, phase: SimPhase, conf: SimConfExportPlotCells) -> None:
    #     '''
    #     Plot the coherence of all cellular microtubules for the cell cluster at
    #     the last time step.
    #     '''
    #
    #     # Prepare to export the microtubules plot.
    #     self._export_prep(phase)
    #
    #     umtx, umty = phase.sim.mtubes.mtubes_to_cell(
    #         phase.cells, phase.p)
    #
    #     pyplot.figure()
    #     ax = pyplot.subplot(111)
    #
    #     plotutil.plotVectField(
    #         umtx,
    #         umty,
    #         phase.cells,
    #         phase.p,
    #         plot_ecm=False,
    #         title='Final Microtubule Alignment Field',
    #         cb_title='Aligned MT Fraction',
    #         colorAutoscale=conf.is_color_autoscaled,
    #         minColor=conf.color_min,
    #         maxColor=conf.color_max,
    #     )
    #
    #     # pyplot.figure()
    #     # ax = pyplot.subplot(111)
    #     # umtx, umty = phase.sim.mtubes.mtubes_to_cell(phase.cells, phase.p)
    #     # plotutil.cell_quiver(umtx, umty, ax, phase.cells, phase.p)
    #
    #     # plotutil.mem_quiver(
    #     #     phase.sim.mtubes.mtubes_x,
    #     #     phase.sim.mtubes.mtubes_y,
    #     #     ax,
    #     #     phase.cells,
    #     #     phase.p,
    #     # )
    #
    #     ax.set_xlabel('X-Distance [um]')
    #     ax.set_ylabel('Y-Distance [um]')
    #     ax.set_title('Net microtubule alignment in cells')
    #
    #     # Export this plot to disk and/or display.
    #     self._export(phase=phase, basename='Final_Microtubules')

    # ..................{ EXPORTERS ~ pump                   }..................
    # @piperunner(
    #     categories=('Ion Pump', 'Density Factor',),
    #     requirements={phasereqs.ELECTROOSMOSIS,},
    # )
    # def export_pump_density(self, conf: SimConfExportPlotCells) -> None:
    #     '''
    #     Plot all cell membrane ion pump density factors for the cell cluster at
    #     the last time step.
    #     '''
    #
    #     # Prepare to export the current plot.
    #     self._export_prep()
    #
    #     plotutil.plotMemData(
    #         phase.cells, phase.p,
    #         zdata=phase.sim.rho_pump,
    #         clrmap=phase.p.default_cm,
    #     )
    #     pyplot.xlabel('Spatial Dimension [um]')
    #     pyplot.ylabel('Spatial Dimension [um]')
    #     pyplot.title('Membrane ion pump density factor')
    #
    #     # Export this plot to disk and/or display.
    #     self._export(basename='final_pumps_2D')

    @piperunner(
        categories=('Ion Pump', 'Pump Rate', 'Na-K-ATPase',),
        requirements=phasereqs.PUMP_NAKATPASE,
    )
    def export_pump_nakatpase(
        self, phase: SimPhase, conf: SimConfExportPlotCells) -> None:
        '''
        Plot all cell membrane Na-K-ATPase pump rates for the cell cluster at
        the last time step.
        '''

        # Prepare to export the current plot.
        self._export_prep(phase)

        pumpData = phase.sim.rate_NaKATP*1e9

        figPump, axPump, cbPump = plotutil.plotPrettyPolyData(
            pumpData,
            phase.sim, phase.cells, phase.p,
            number_cells=phase.p.visual.is_show_cell_indices,
            clrmap=phase.p.default_cm,
        )

        axPump.set_title('Final Na/K-ATPase Pump Rate')
        axPump.set_xlabel('Spatial distance [um]')
        axPump.set_ylabel('Spatial distance [um]')
        cbPump.set_label('Pump Na+ Flux (nmol/m2*s)')

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='final_NaKPump_2D')

    # ..................{ EXPORTERS ~ pressure               }..................
    @piperunner(
        categories=('Pressure', 'Total',),
        requirements=phasereqs.PRESSURE_TOTAL,
    )
    def export_pressure_total(
        self, phase: SimPhase, conf: SimConfExportPlotCells) -> None:
        '''
        Plot all **cellular pressure totals** (i.e., summations of all cellular
        mechanical and osmotic pressures) for the cell cluster at the last time
        step.
        '''

        # Prepare to export the current plot.
        self._export_prep(phase)

        figP, axP, cbP = plotutil.plotPolyData(
            phase.sim, phase.cells, phase.p,
            zdata=phase.sim.P_cells,
            number_cells=phase.p.visual.is_show_cell_indices,
            clrmap=phase.p.default_cm,
            clrAutoscale=conf.is_color_autoscaled,
            clrMin=conf.color_min,
            clrMax=conf.color_max,
        )

        axP.set_title('Final Pressure in Cell Network')
        axP.set_xlabel('Spatial distance [um]')
        axP.set_ylabel('Spatial distance [um]')
        cbP.set_label('Pressure [Pa]')

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='final_P_2D_gj')

    # ..................{ EXPORTERS ~ voltage                }..................
    @piperunner(
        categories=('Voltage', 'Extracellular',),
        requirements=phasereqs.VOLTAGE_EXTRA,
    )
    def export_voltage_extra(
        self, phase: SimPhase, conf: SimConfExportPlotCells) -> None:
        '''
        Plot all extracellular voltages for the environment at the last time
        step.
        '''

        # Prepare to export the current plot.
        self._export_prep(phase)

        vv = phase.sim.v_env.reshape(phase.cells.X.shape)

        pyplot.figure()
        pyplot.imshow(
            1e3*vv,
            origin='lower',
            extent=phase.cache.upscaled.extent,
            cmap=phase.p.default_cm,
        )
        pyplot.colorbar()
        pyplot.title('Environmental Voltage [mV]')

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='Final_environmental_V')


    @piperunner(
        categories=('Voltage', 'Polarity',),
        requirements=phasereqs.VOLTAGE_POLARITY,
    )
    def export_voltage_polarity(
        self, phase: SimPhase, conf: SimConfExportPlotCells) -> None:
        '''
        Plot all cellular voltage polarities for the cell cluster at the last
        time step.
        '''

        # Prepare to export the polarization plot.
        self._export_prep(phase)

        pyplot.figure()

        # Plot a background Vmem mesh.
        fig, ax, cb = plotutil.plotPrettyPolyData(
            1000*phase.sim.vm_time[-1],
            phase.sim, phase.cells, phase.p,
            number_cells=phase.p.visual.is_show_cell_indices,
            current_overlay=False,
            plotIecm=phase.p.IecmPlot,
            clrmap=phase.p.default_cm,
            clrAutoscale=conf.is_color_autoscaled,
            clrMin=conf.color_min,
            clrMax=conf.color_max,
        )

        # Calculate the Vmem polarity vectors.
        polm = phase.sim.vm - (
            phase.sim.vm_ave_time[-1][phase.cells.mem_to_cells])
        polx = polm*phase.cells.mem_vects_flat[:,2]
        poly = polm*phase.cells.mem_vects_flat[:,3]

        pcx = np.dot(
            phase.cells.M_sum_mems,
            polx*phase.cells.mem_sa) / phase.cells.cell_sa
        pcy = np.dot(
            phase.cells.M_sum_mems,
            poly*phase.cells.mem_sa) / phase.cells.cell_sa

        plotutil.cell_quiver(pcx, pcy, ax, phase.cells, phase.p)

        ax.set_xlabel('X-Distance [um]')
        ax.set_ylabel('Y-Distance [um]')
        ax.set_title('Cell Vmem polarity')

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='Final_Polarity')

    # ..................{ EXPORTERS ~ voltage : vmem         }..................
    @piperunner(categories=('Voltage', 'Transmembrane', 'Actual',))
    def export_voltage_membrane(
        self, phase: SimPhase, conf: SimConfExportPlotCells) -> None:
        '''
        Plot all transmembrane voltages (Vmem) for the cell cluster at the last
        time step.
        '''

        # Prepare to export the current plot.
        self._export_prep(phase)

        figV, axV, cbV = plotutil.plotPrettyPolyData(
            1000*phase.sim.vm_time[-1],
            phase.sim, phase.cells, phase.p,
            number_cells=phase.p.visual.is_show_cell_indices,
            current_overlay=False,
            plotIecm=phase.p.IecmPlot,
            clrmap=phase.p.default_cm,
            clrAutoscale=conf.is_color_autoscaled,
            clrMin=conf.color_min,
            clrMax=conf.color_max,
        )

        figV.suptitle('Final Vmem', fontsize=14, fontweight='bold')
        axV.set_xlabel('Spatial distance [um]')
        axV.set_ylabel('Spatial distance [um]')
        cbV.set_label('Voltage mV')

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='final_Vmem_2D')


    @piperunner(categories=('Voltage', 'Transmembrane', 'Average',))
    def export_voltage_membrane_average(
        self, phase: SimPhase, conf: SimConfExportPlotCells) -> None:
        '''
        Plot the averages of all transmembrane voltages (Vmem) for the cell
        cluster at the last time step.
        '''

        # Prepare to export the current plot.
        self._export_prep(phase)

        figVa, axVa, cbVa = plotutil.plotPolyData(
            phase.sim, phase.cells, phase.p,
            zdata=1000*phase.sim.vm_ave,
            number_cells=phase.p.visual.is_show_cell_indices,
            current_overlay=False,
            plotIecm=phase.p.IecmPlot,
            clrmap=phase.p.default_cm,
            clrAutoscale=conf.is_color_autoscaled,
            clrMin=conf.color_min,
            clrMax=conf.color_max,
        )

        # axVa.quiver(
        #     p.um*cells.cell_centres[:,0],
        #     p.um*cells.cell_centres[:,1], sim.pol_cell_x, sim.pol_cell_y)

        figVa.suptitle('Final Average Vmem', fontsize=14, fontweight='bold')
        axVa.set_xlabel('Spatial distance [um]')
        axVa.set_ylabel('Spatial distance [um]')
        cbVa.set_label('Voltage [mV]')

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='final_AverageVmem_2D')


    @piperunner(
        categories=('Voltage', 'Transmembrane', 'GHK',),
        requirements=phasereqs.VOLTAGE_MEMBRANE_GHK,
    )
    def export_voltage_membrane_ghk(
        self, phase: SimPhase, conf: SimConfExportPlotCells) -> None:
        '''
        Plot all transmembrane voltages (Vmem) calculated by the
        Goldman-Hodgkin-Katz (GHK) equation for the cell cluster at the last
        time step.
        '''

        # Prepare to export the current plot.
        self._export_prep(phase)

        figV_ghk, axV_ghk, cbV_ghk = plotutil.plotPolyData(
            phase.sim, phase.cells, phase.p,
            zdata=1000*phase.sim.vm_GHK_time[-1],
            number_cells=phase.p.visual.is_show_cell_indices,
            current_overlay=False,
            plotIecm=phase.p.IecmPlot,
            clrmap=phase.p.default_cm,
            clrAutoscale=conf.is_color_autoscaled,
            clrMin=conf.color_min,
            clrMax=conf.color_max,
        )

        figV_ghk.suptitle(
            'Final Vmem using Goldman Equation',
            fontsize=14,
            fontweight='bold',
        )
        axV_ghk.set_xlabel('Spatial distance [um]')
        axV_ghk.set_ylabel('Spatial distance [um]')
        cbV_ghk.set_label('Voltage [mV]')

        # Export this plot to disk and/or display.
        self._export(phase=phase, basename='final_Vmem_GHK_2D')



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

# ....................{ IMPORTS                            }....................
import numpy as np
from betse.lib.pil import pilnumpy
from betse.lib.pil.pilnumpy import ImageModeType
from betse.science.math import finitediff as fd
from betse.util.math.mathinterp import interp2d_linear
from betse.util.path import pathnames
# from scipy.ndimage import gaussian_filter

# ....................{ IMPORTS                            }....................
def periodic(pc,cells,p):
    """
    Creates a time-dependent signal of a sinusoidal function.

    Parameters
    ----------
    pc                 Indices of cells (p.is_ecm == False) or membranes (p.is_ecm == True)
    cells              Instance of Cells module
    p                  Instance of parameters

    Returns
    ---------
    y              Data values from 0 to 1 defining a periodic signal.
    scalar         Null output corresponding to spatial variation.
    """

    scalar = 1

    y = lambda t: np.sin(
        t*np.pi*(p.periodic_properties['frequency']/1) +
        p.periodic_properties['phase']
    ) ** 2

    return scalar, y


def f_sweep(pc, cells, p) -> tuple:
    '''
    Create and return a time-dependent sweep of a sinusoidal function through
    various frequencies.

    Parameters
    ----------
    pc                 Indices of cells (p.is_ecm == False) or membranes (p.is_ecm == True)
    cells              Instance of Cells module
    p                  Instance of parameters

    Returns
    ---------
    y              Data values from 0 to 1 defining a periodic signal.
    scalar         Null output corresponding to spatial variation.
    '''

    scalar = 1

    if p.f_scan_properties['f slope'] is None and p._run_sim:
        p.f_scan_properties['f slope'] = (
            p.f_scan_properties['f stop'] -
            p.f_scan_properties['f start']) / p.sim_time_total

    f_slope = p.f_scan_properties['f slope']

    y = lambda t: np.sin(
        np.pi*(f_slope*t + p.f_scan_properties['f start'])*t)**2

    return scalar, y

def single_cell(pc, cells, p):
    '''
    Randomly selects a single cell on the cluster boundary and generates a modulator
    function with a max value at the single cell and zero value elsewhere (with offset)

    Parameters
    ----------
    pc                 Indices of membranes or cells
    cells              Instance of Cells module
    p                  Instance of parameters
    '''

    # Randomly choose a boundary cell:
    cell_ind = np.random.choice(cells.bflags_cells)

    # Create an array of zeros with length equal to number of cell centres and add the
    # single cell index as the only value:
    mod_funk_cells = np.zeros(len(cells.cell_i))
    mod_funk_cells[cell_ind] = 1.0

    if len(pc) == len(cells.mem_i):
        # Let's broadcast it out to membranes of the randomly chosen cell:
        mod_funk = mod_funk_cells[cells.mem_to_cells] + p.mod_single_cell_offset
    elif len(pc) == len(cells.cell_i):
        # Don't broadcast; keep as single cell:
        mod_funk = mod_funk_cells + p.mod_single_cell_offset

    dynamics = lambda t: 1

    return mod_funk, dynamics

def gradient_x(pc, cells,p):
    """
    Creates a spatial gradient along the x-axis from 0 to 1 over a patch
    of cells defined by indices.

    Parameters
    ----------
    pc                 Indices of membranes or cells
    cells              Instance of Cells module
    p                  Instance of parameters
    targs              Indices of targets within mem or cell inds to apply gradient

    Returns
    ---------
    fx
        Data values from 0 to 1 defining a gradient in the x-direciton. If
        p.is_ecm == True, length of fx is number of membranes, otherwise length
        of fx is equal to cell number.
    dynamics
        Null quantity corresponding to time dynamics.
    """

    if len(pc) == len(cells.mem_i):
        x_vals_o = cells.mem_mids_flat[:, 0]
    elif len(pc) == len(cells.cell_i):
        x_vals_o = cells.cell_centres[:, 0]

    min_grad = x_vals_o.min()

    # shift values to have a minimum of zero:
    x_vals = x_vals_o - min_grad

    # scale values to have a maximum of 1:
    x_vals = x_vals/x_vals.max()

    # shift the x_mid value by the offset:
    x_mid = 0.5 + p.gradient_x_properties['x-offset']

    n = p.gradient_x_properties['exponent']
    grad_slope = ((x_vals/x_mid)**n)/(1+((x_vals/x_mid)**n))
    fx = (grad_slope)*p.gradient_x_properties['slope'] + p.gradient_x_properties['z-offset']

    dynamics = lambda t: 1

    return fx, dynamics

def gradient_y(pc, cells,p):
    """
    Creates a spatial gradient along the y-axis from 0 to 1 over a patch
    of cells defined by indices.

    If p.is_ecm is True, the data is defined and returned on membrane
    midpoints, otherwise cell centres are the defining structure.

    Parameters
    ----------
    pc                 Indices of membranes
    cells              Instance of Cells module
    p                  Instance of parameters

    Returns
    ---------
    fy
        Data values from 0 to 1 defining a gradient in the y-direciton. If
        p.is_ecm == True, length of fy is number of membranes, otherwise length
        of fy is equal to cell number.
    dynamics
        Null quantity corresponding to time dynamics.
    """

    if len(pc) == len(cells.mem_i):
        y_vals_o = cells.mem_mids_flat[:, 1]
    elif len(pc) == len(cells.cell_i):
        y_vals_o = cells.cell_centres[:, 1]

    min_grad = y_vals_o.min()

    # shift values to have a minimum of zero:
    y_vals = y_vals_o - min_grad

    # scale values to have a maximum of 1:
    y_vals = y_vals / y_vals.max()

    # shift the y_mid value by the offset:
    y_mid = 0.5 + p.gradient_y_properties['x-offset']

    n = p.gradient_y_properties['exponent']
    grad_slope = ((y_vals / y_mid) ** n) / (1 + ((y_vals / y_mid) ** n))
    fy = (grad_slope) * p.gradient_y_properties['slope'] + p.gradient_y_properties['z-offset']

    dynamics = lambda t: 1

    return fy, dynamics

def gradient_r(pc, cells,p):
    """
    Creates a spatial gradient in a radial direction from 0 to 1 over a patch
    of cells defined by indices.

    If p.is_ecm is True, the data is defined and returned on membrane
    midpoints, otherwise cell centres are the defining structure.

    Parameters
    ----------
    pc                 Indices of membranes
    cells              Instance of Cells module
    p                  Instance of parameters

    Returns
    ---------
    r
        Data values from 0 to 1 defining a gradient in the radial direction.  If
        p.is_ecm == True, length of r is number of membranes, otherwise length
        of r is equal to cell number.
    """

    fx = np.zeros(len(cells.mem_i))
    fy = np.zeros(len(cells.mem_i))

    fx[pc] = (
        cells.mem_mids_flat[pc,0] -
        cells.centre[0] -
        p.gradient_r_properties['x-offset'])
    fy[pc] = (
        cells.mem_mids_flat[pc,1] -
        cells.centre[1] -
        p.gradient_r_properties['x-offset'])

    r = np.sqrt(fx**2 + fy**2)

    r = r/r.max()

    r = r*p.gradient_r_properties['slope'] + p.gradient_r_properties['z-offset']

    dynamics = lambda t: 1

    return r, dynamics

def gradient_bitmap(pc, cells, p, bitmap_filename = None):

    """
    This modulator reads in a bitmap supplied by the user from the
    directory in params.

    Red is treated as positive, blue is treated as negative, and
    black is treated as zero.

    Parameters
    ------------

    pc:   array of cell centre or membrane midpoints
    cells:  BETSE cells object
    p:      BETSE parameters object

    """

    if bitmap_filename is None:  # default this to loading the 'gradient bitmap function' from params
        bitmap_filename = p.grad_bm_fn
        grad_bm_offset = np.max((p.grad_bm_offset, 0.0))

    else:
        grad_bm_offset = 0.0


    if len(pc) == len(cells.mem_i):
        xmap = cells.map_mem2ecm

    elif len(pc) == len(cells.cell_i):
        xmap = cells.map_cell2ecm

    # xmi = cells.xmin - 4*p.cell_radius
    # ymi = cells.ymin - 4*p.cell_radius
    # xma = cells.xmax + 4*p.cell_radius
    # yma = cells.ymax + 4*p.cell_radius

    xmi = cells.xmin
    ymi = cells.ymin
    xma = cells.xmax
    yma = cells.ymax

    xx = np.linspace(cells.xmin, cells.xmax, cells.X.shape[1])
    yy = np.linspace(cells.ymin, cells.ymax, cells.X.shape[0])

    fn1 = pathnames.join(p.conf_dirname, bitmap_filename)

    # Three-dimensional Numpy array of the RGB-ordered integer components of all
    # pixels loaded from this image.
    a1o = pilnumpy.load_image(filename=fn1, mode=ImageModeType.COLOR_RGB)

    a1 = np.asarray(a1o, dtype=np.float64)

    a1_F = (a1[:, :, 0] -  a1[:, :, 2]) / 255

    a1_F = np.flipud(a1_F)
    # a1_F = fd.integrator(a1_F, sharp=0.5) # smooth a little to avoid bizarre visual effects

    xa = np.linspace(xmi, xma, a1_F.shape[1])
    ya = np.linspace(ymi, yma, a1_F.shape[0])

    spline_F = interp2d_linear(xa, ya, a1_F)
    # spline_F = interpolate.interp2d(xa, ya, a1_F, kind='linear', fill_value=0.0)
    fe = spline_F(xx, yy)

    fe = fd.integrator(fe, sharp=0.5) # smooth a little to avoid bizarre visual effects

    f = fe.ravel()[xmap]

    # f = (f/f.max()) + grad_bm_offset
    f = f + grad_bm_offset

    # indz = (f < 0.0).nonzero()
    # f[indz] = 0.0
    dynamics = lambda t:1

    return f, dynamics



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level vector subcache functionality.
'''

# ....................{ IMPORTS                            }....................
from betse.science.math.cache.cacheabc import SimPhaseCacheABC
from betse.science.visual.layer.vector import lyrvecabc
from betse.science.visual.layer.vector.lyrvecabc import (
    LayerCellsVectorColorfulABC)
from betse.util.type.decorator.decmemo import property_cached

# ....................{ SUBCLASSES                         }....................
class SimPhaseCacheLayerCellsVector(SimPhaseCacheABC):
    '''
    Simulation phase-specific **vector layer** (i.e., vector-based layer
    layering a one-dimensional Numpy array onto exported visuals) subcache,
    persisting all previously constructed vector-based layers for a single
    simulation phase.
    '''

    # ..................{ PROPERTIES ~ currents              }..................
    @property_cached
    def voltage_membrane(self) -> LayerCellsVectorColorfulABC:
        '''
        Vector layer layering all transmembrane voltages (Vmem) for the cell
        cluster over all time steps onto caller-provided visuals.
        '''

        return lyrvecabc.make_layer(
            phase=self._phase, vector=self._phase.cache.vector.voltage_membrane)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level **application dependency metadata** (i.e., lists of version-pinned
dependencies synopsizing application requirements) functionality.
'''

# ....................{ IMPORTS                           }....................
# from betse.util.io.log import logs
from betse.util.type.types import (
    type_check, MappingType, ModuleType, IterableTypes)

# ....................{ GLOBALS                           }....................
MERGE_MODULE_METADEPS_DICTS_NAME = (
    'RUNTIME_MANDATORY',
    'RUNTIME_OPTIONAL',
    'TESTING_MANDATORY',
    'REQUIREMENT_NAME_TO_COMMANDS',
)
'''
Tuple of the names of all global dictionaries required to be defined by each
application dependency metadata module passed as input to the
:func:`merge_module_metadeps` function via the ``modules_metadeps`` parameter.
'''

# ....................{ MAKERS                            }....................
#FIXME: Improve documentation with respect to key collisions. Specifically,
#provide concrete examples without code (which would probably be a bit too
#much, frankly) of how this function resolves the following similar edge cases:
#
#* Two or more dictionaries to be merged contain key-value pairs containing the
#  same key but differing values.
#* Two or more dictionaries to be merged contain the same key-value pairs.

@type_check
def merge_module_metadeps(
    module_name: str, modules_metadeps: IterableTypes) -> ModuleType:
    '''
    Dynamically create and return a new **application dependency metadata
    module** (i.e., module satisfying the informal protocol defined by the
    :mod:`betse.metadeps` module) with the passed name, iteratively merging the
    contents of all passed application dependency metadata modules.

    Specifically, this function (in order):

    #. Dynamically creates a new module with the passed name.
    #. For each of the global attributes assumed by the application dependency
       metadata protocol (i.e., dictionaries named ``RUNTIME_MANDATORY``,
       ``RUNTIME_OPTIONAL``, ``TESTING_MANDATORY``, and
       ``REQUIREMENT_NAME_TO_COMMANDS``), creates a new global attribute of the
       same name in this new module whose value is a dictionary merging the
       contents of all dictionaries of the same name defined by all passed
       modules. For example, if passed the :mod:`betse.metadeps` and
       :mod:`betsee.guimetadeps` modules, this function creates a new module
       defining a global dictionary named ``RUNTIME_MANDATORY`` merging the
       contents of the :attr:`betse.metadeps.RUNTIME_MANDATORY` and
       :mod:`betsee.guimetadeps.RUNTIME_MANDATORY` dictionaries.
    #. Returns this new module.

    Motivation
    ----------
    This utility function is *only* intended to be called by subclass
    implementations of the abstract private
    :meth:`betse.util.app.meta.appmetaabc.AppMetaABC._module_metadeps` property
    in downstream consumers. Notably, BETSEE calls this function from its
    concrete implementation of this property to merge its own dependency
    requirements with those of BETSE; doing so guarantees that calls to
    dependency-based functions within the BETSE codebase (e.g., of the
    :func:`betse.lib.libs.import_runtime_optional` function dynamically
    importing one or more optional dependencies) behave as expected.

    Caveats
    ----------
    **Order is insignificant.** If any of the requisite global dictionaries
    defined by any of the passed modules contain one or more key-value pairs
    contained in any other dictionary of the same name defined by any other
    passed module, this function raises an exception. Since this prevents any
    collisions between key-value pairs, *no* implicit precedence exists between
    these modules.

    Parameters
    ----------
    module_name : str
        Fully-qualified name of the module to be created.
    modules_metadeps : IterableTypes
        Iterable of all application dependency metadata modules to be merged.
        Order is insignificant. See above.

    Raises
    ----------
    BetseAttrException
        If any of these modules fail to define a requisite attribute.
    BetseMappingException
        If any two global dictionaries of the same name defined by any two of
        these modules **item-collide** (i.e., if any key-value pair in any such
        dictionary is also a key-value pair in any other such dictionary).
    BetseModuleException
        If a module with this target module name already exists.
    BetseTypeException
        If any of the requisite attributes defined by any of these modules are
        *not* dictionaries.

    Returns
    ----------
    ModuleType
        Output application dependency metadata module with this name, merging
        the contents of these input application dependency metadata modules.
    '''

    # Avoid circular import dependencies.
    from betse.util.py.module import pymodname, pymodule
    from betse.util.type.iterable import itertest, sequences
    from betse.util.type.iterable.mapping import mapmerge
    from betse.util.type.obj import objects
    from betse.util.type.text.string import strjoin

    # Sequence of modules converted from this iterable of modules.
    modules_metadeps = sequences.to_sequence(modules_metadeps)

    # If less than two modules were passed, raise an exception.
    sequences.die_if_length_less_than(sequence=modules_metadeps, length=2)
    # Else, at least two modules were passed.

    # If any of the passed modules is *NOT* a module, raise an exception.
    itertest.die_unless_items_instance_of(
        iterable=modules_metadeps, cls=ModuleType)

    # Dictionary mapping from the name to value of each module-scoped
    # attribute to be declared in the module to be created and returned,
    # defaulting to the "RequirementCommand" class globally defined by the
    # first such module.
    trg_module_attr_name_to_value = {
        'RequirementCommand': objects.get_attr(
            obj=modules_metadeps[0], attr_name='RequirementCommand'),
    }

    # For the name of each such global dictionary...
    for module_dict_name in MERGE_MODULE_METADEPS_DICTS_NAME:
        # Generator comprehension aggregating all of the global dictionaries
        # defined by all of these input modules, raising exceptions if any such
        # module fails to define such a dictionary.
        src_modules_dict = (
            objects.get_attr(
                obj=modules_metadep,
                attr_name=module_dict_name,
                attr_type=MappingType,
            )
            for modules_metadep in modules_metadeps
        )

        # Merge these dictionaries into the dictionary to be returned, raising
        # exceptions if any requirement defined by any such dictionary
        # item-collides (i.e., if any key-value pair in any such dictionary is
        # also a key-value pair in any other such dictionary).
        trg_module_attr_name_to_value[module_dict_name] = mapmerge.merge_maps(
            src_modules_dict)

    # Double-quoted conjunction of the fully-qualified names of all passed
    # input application dependency metadata modules.
    src_modules_metadeps_name = (
        strjoin.join_iterable_as_conjunction_double_quoted(
            pymodule.get_name_qualified(module_metadep)
            for module_metadep in modules_metadeps
        ))

    # Output application dependency metadata module to be returned.
    trg_module_metadeps = pymodname.make_module(
        module_name=module_name,
        module_attr_name_to_value=trg_module_attr_name_to_value,
        module_doc='''
**Application dependency metadata** (i.e., sequences of version-pinned
dependencies synopsizing application requirements), dynamically merged from the
{} modules.'''.format(src_modules_metadeps_name),
    )

    # Return this module.
    return trg_module_metadeps



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level shell-specific string facilities.
'''

# ....................{ IMPORTS                           }....................
from betse.util.type.types import type_check, IntOrNoneTypes

# ....................{ GETTERS                           }....................
#FIXME: Consider integrating this method with the main codebase, ideally by
#calling this function before running each simulation phase. This banner is
#sufficiently aesthetic that I'd certainly appreciate seeing it everywhere.
@type_check
def get_banner(
    # Mandatory arguments.
    title: str,

    # Optional arguments.
    padding: str = '=',
    length: IntOrNoneTypes = None,
) -> str:
    '''
    Single-line banner embedding the passed human-readable title.

    Parameters
    ----------
    title : str
        Human-readable substring to be embedded in this banner, ideally less
        than or equal to 40 characters in length (i.e., half the default
        terminal width and thus banner length of 80 characters).
    padding : optional[str]
        Single character with which to pad this title. To visually center this
        title in this banner, this character is repeated both before and after
        this title as necessary. Defaults to an ASCII punctuation character.
    length : IntOrNoneTypes
        Preferred length of the returned banner. If the length of this title
        does *not* already exceed this preferred length, the returned banner is
        guaranteed to have this exact length. Defaults to ``None``, in which
        case this length internally defaults to:
        * If the current shell is an interactive terminal, the **width** (i.e.,
          maximum number of columns currently displayed by) this terminal.
        * Else, the conventional default of 80.

    Returns
    ----------
    str
        Single-line banner embedding this title.

    Examples
    ----------
        >>> from betse.util.os.shell import shellstr
        >>> shellstr.get_banner(
        ...     title='The Love Song of J. Alfred Prufrock', padding='~')
        ~~~~~~~~~~~~~~~~~~~~~~ The Love Song of J. Alfred Prufrock ~~~~~~~~~~~~~~~~~~~~~
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text import chars

    # If this padding is *NOT* a single character, raise an exception.
    chars.die_unless_char(padding)

    #FIXME: Generalize this hard-coded maximum length as follows:
    #
    #* If the current shell is interactive, set this length to this shell's
    #  current column length. Sadly, doing so requires platform-specific logic.
    #* Else, default this length to the current reasonable default of 80.

    # If the caller passed no preferred banner length...
    if length is None:
        # Default this length to the conventional terminal width.
        length = 80

    # Minimal-length banner containing this text and the smallest
    # possible prefixing and suffixing banner characters.
    banner_min = '{padding} {title} {padding}'.format(
        title=title, padding=padding)

    # If the length of this minimal-length banner exceeds the maximum
    # length, print this text with no such banner.
    if len(banner_min) > length:
        return title
    # Else, print this text within a banner.
    else:
        #FIXME: Again, generalize this hard-coded length in a manner dependent
        #on the current terminal width.

        # Length of the string preceding this text. To produce a uniform
        # rather than ragged left margin for aesthetic uniformity, this
        # length is a constant.
        BANNER_PREFIX_LEN = 30

        # String preceding this text.
        banner_prefix = padding * BANNER_PREFIX_LEN

        # Banner to be printed excluding all suffixing banner characters.
        banner_sans_suffix = '{} {} '.format(banner_prefix, title)

        # Length of the string following this text. To dynamically fill all
        # remaining line space, this length contextually depends on the
        # lengths of all other strings. By the above conditional, this
        # length is guaranteed to be non-zero and need *NOT* be tested.
        banner_suffix_len = length - len(banner_sans_suffix)

        # Nonetheless, test this length for sanity.
        assert banner_suffix_len > 0, (
            'Banner suffix length {} not positive.'.format(
                banner_suffix_len))

        # String following this text.
        banner_suffix = padding * banner_suffix_len

        # Single-line banner to be returned.
        banner = '\n{}{}'.format(banner_sans_suffix, banner_suffix)

        # Return this banner.
        return banner

# ....................{ QUOTERS                           }....................
@type_check
def shell_quote(text: str) -> str:
    '''
    Shell-quote the passed string in a platform-specific manner.

    If the current platform is:

    * *Not* Windows (e.g., Linux, OS X), the returned string is guaranteed to
      be suitable for passing as an arbitrary positional argument to external
      commands.
    * Windows, the returned string is suitable for passing *only* to external
      commands parsing arguments in the same manner as the Microsoft C runtime.
      While *all* applications on POSIX-compliant systems are required to parse
      arguments in the same manner (i.e., according to Bourne shell lexing), no
      such standard applies to Windows applications. Shell quoting is therefore
      fragile under Windows -- like pretty much everything.
    '''

    # Avoid circular import dependencies.
    from betse.util.os.brand import windows

    # If the current OS is Windows, do *NOT* perform POSIX-compatible quoting.
    # Windows is POSIX-incompatible and hence does *NOT* parse command-line
    # arguments according to POSIX standards. In particular, Windows does *NOT*
    # treat single-quoted arguments as single arguments but rather as multiple
    # shell words delimited by the raw literal `'`. This is circumventable by
    # calling an officially undocumented Windows-specific function. (Awesome.)
    if windows.is_windows():
        import subprocess
        return subprocess.list2cmdline([text])
    # Else, perform POSIX-compatible quoting.
    else:
        import shlex
        return shlex.quote(text)
